#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include "nrutils.h"

#define SWAP(a,b) tempr=(a);(a) = (b);(b)=tempr


/*
 * Replaces data[1..2*nn] by its discrete Fourier transform, if 
 * isign is input as 1; or replaces data[1..2*nn] by nn times its 
 * inverse discrete Fourier transform, if isign is input as -1.
 * Data is a complex array of length nn or, equivalently, a real 
 * array of length 2*nn. nn MUST be an integral power of 2 (unchecked).
 */
void four1(data, nn, isign) 
     float data[];
     unsigned long nn;
     int isign;
{
  unsigned long n, mmax, m, j, istep, i;
  /* double precision for trigonometric recurrences */
  double wtemp, wr, wpr, wpi, wi, theta;
  float tempr, tempi;

#ifdef DEBUG
  printf("In four1 > data=%x, nn=%d, isign=%d ...\n", data,nn,isign);
#endif

  n = nn << 1;
  j = 1;
  for (i=1; i < n; i += 2) {   /* This is the bit-reversal section of */
    if (j > i) {               /*    the routine */
      SWAP(data[j], data[i]);  /* Exchange the two complex numbers */
      SWAP(data[j+1], data[i+1]);
    }
    m = n >> 1;
    while (m >=2 && j > m) {
      j -= m;
      m >>= 1;
    }
    j += m;
  }
#ifdef DEBUG
 printf("In four1 > Beginning Danielson-Lanczos section ...\n");
#endif

  /* 
   * Here begins the Danielson-Lanczos section of the routine
   */
  mmax = 2;
  while (n > mmax) {       /* Outer loop executed log(nn) times */
    istep = mmax << 1;
    theta = isign * (6.28318530717959/mmax);  /* init trig recurrence */
    wtemp = sin(0.5 * theta);
    wpr = -2.0 * wtemp * wtemp;
    wpi = sin(theta);
    wr = 1.0;
    wi = 0.0;
    for (m=1; m < mmax; m += 2) {       /* The two nested loops */
      for (i=m; i <= n; i += istep) {
	/* 
	 * This is the Danielson-Lanczos formula:
	 */
	j = i+mmax;
	tempr = wr * data[j] - wi*data[j+1];
	tempi = wr * data[j+1] + wi*data[j];
	data[j] = data[i] - tempr;
	data[j+1] = data[i+1] - tempi;
	data[i] += tempr;
	data[i+1] += tempi;
      }
      wr = (wtemp = wr)*wpr - wr*wpi + wr; /* trigonometric recurrence */
      wi = wi*wpr + wtemp*wpi + wi;
    }
    mmax = istep;
  }
#ifdef DEBUG
 printf("In four1 > About to exit four1 ...\n");
#endif
}
 
/*
 * Calculates the Fourier transform of a set of n real-valued data
 * points. Replaces this data (which is stored in array data[1..n])
 * by the positive frequency half of its complex Fourier transform.
 * The real-valued first and last components of the complex transform
 * are returned as elements data[1] and data[2], respectively. n must
 * be a power of 2. This routine also calculates the inverse transform
 * of a complex data array if it is the transform of real data. 
 * (Result in this case must be multiplied by 2/n.)
 */
void realft(data, n, isign) 
     float data[];
     unsigned long n;
     int isign;
{
  unsigned long i, i1, i2, i3, i4, np3;
  float c1 = 0.5;
  float c2, h1r, h1i, h2r, h2i;
  double wr, wi, wpr, wpi, wtemp, theta;

#ifdef DEBUG
 printf("In realft > data=%x, n=%d, isign=%d ...\n",data,n,isign);
#endif

  theta = 3.141592653589793/(double) (n>>1);   /* init trig recurrence */
  if (isign == 1) {
    c2 = -0.5;
    four1(data, n >> 1, 1);                    /* forward transform */
  } else {
    c2 = 0.5;
    theta = -theta;                 /* set up for inverse transform */
  }

  wtemp = sin(0.5 * theta);
  wpr = -2.0 * wtemp * wtemp;
  wpi = sin(theta);
  wr = 1.0 + wpr;
  wi = wpi;
  np3 = n+ 3;

#ifdef DEBUG
 printf("In realft > Separation/recombination of transform ... \n");
#endif
  for (i=2; i <= (n >> 2); i++) {   /* i=1 done below */
    i4 = 1 + (i3 = np3 - (i2 = 1 + (i1 = i + i - 1)));
    /* 
     * The two serparate transforms are separated out of data
     */
    h1r = c1*(data[i1] + data[i3]);
    h1i = c1*(data[i2] - data[i4]);
    h2r = -c2*(data[i2] + data[i4]);
    h2i = c2*(data[i1] - data[i3]);
    /* 
     * Here they are recombined to form the true transform
     * of the original real data.
     */
    data[i1] = h1r + wr*h2r - wi*h2i;
    data[i2] = h1i + wr*h2i + wi*h2r;
    data[i3] = h1r - wr*h2r + wi*h2i;
    data[i4] = -h1i + wr*h2i + wi*h2r;
    wr = (wtemp = wr)*wpr - wr*wpi + wr;  /* The recurrence */
    wi = wi*wpr + wtemp*wpr + wi;
  }
  if (isign == 1) {
    /*
     * Squeeze the first and last data together
     * to get them all within the original array.
     */
    data[1] = (h1r = data[1]) + data[2];
    data[2] = h1r - data[2];
  } else {
    data[1] = c1 * ((h1r = data[1]) + data[2]);
    data[2] = c1 * (h1r - data[2]);
#ifdef DEBUG
 printf("In realft > Calling four1() ...\n");
#endif
    four1(data, n >> 1, -1);       /* The inverse transform for the */
  }				                   /* case isign = -1. */
#ifdef DEBUG
 printf("In realft > Exiting realft ...\n");
#endif
}

/*
 * Given two real input arrays data1[1..n] and data2[1..n], this
 * routine calls four1 and returns two complex output arrays,
 * fft1[1..2n] and fft2[1..2n], each of complex length n (i.e., 
 * real length 2*n), which contain the discrete Fourier transforms 
 * of the respective data arrays. n MUST be an integral power of 2.
 */
void twofft(data1, data2, fft1, fft2, n) 
   float data1[];
   float data2[];
   float fft1[];
   float fft2[];
   unsigned long n;
{
   unsigned long nn3, nn2, jj,j;
   float rep, rem, aip, aim;

#ifdef DEBUG
 printf("In twofft > data1=%x, data2=%x, fft1=%x, fft2=%x, n=%d ...\n",
        data1, data2, fft1, fft2, n);
#endif

   nn3 = 1+(nn2 = 2+n+n);
   /*
    * Pack the two real arrays into one complex array.
    */
   for (j=1, jj=2; j <= n; j++, jj+=2) {  
     fft1[jj-1] = data1[j];
     fft1[jj] = data2[j];
   }

#ifdef DEBUG
 printf("In twofft > Calling four1() ...\n");
#endif
   four1(fft1, n, 1);   /* Transform the complex array */
   fft2[1] = fft1[2];
   fft1[2] = fft2[2] = 0.0;
#ifdef DEBUG
 printf("In twofft > Setting up complex arrays\n");
#endif
   for (j=3; j <= n+1; j+=2) {
     /* 
      * use symmetries to separate the two transforms 
      */
     rep = 0.5 * (fft1[j] + fft1[nn2-j]);
     rem = 0.5 * (fft1[j] - fft1[nn2-j]);
     aip = 0.5 * (fft1[j+1] + fft1[nn3-j]);
     aim = 0.5 * (fft1[j+1] - fft1[nn3-j]);
     fft1[j] = rep;     /* Ship them out in two complex arrays. */
     fft1[j+1] = aim;
     fft1[nn2-j] = rep;
     fft1[nn3-j] = -aim;
     fft2[j] = aip;
     fft2[j+1] = -rem;
     fft2[nn2-j] = aip;
     fft2[nn3-j] = rem;
  }

  for (j=1; j <= n; j++) {
     printf("%.4f %.4f\n", fft1[j], fft1[j+1]);   /*#########*/
  }
#ifdef DEBUG
 printf("In twofft > exiting twofft() ...\n");
#endif
}

/*
 * Convolves or deconvolves a real data set data[1..n] (including 
 * any uer-supplied zero padding) with aresponse function respns[1..n].
 * The response function must be stored in wrap-around order in the
 * first m elements of respns, where m is an odd integer <= n. Wrap-
 * around order means that the first half of the array respns contains
 * the impulse response function at positive times, while the second
 * half of the array contains the impulse response function at negative
 * times, counting down fromthe highest element respns[m]. On input 
 * isign is +1 for convolution, -1 for deconvolution. The answer is 
 * returned in the first n components of ans. However ans must be 
 * supplied in the calling program with dimensions [1..2*n], for 
 * consistency with twofft. n MUST be an integer power of 2.
 */
void convlv(data, n, respns, m, isign, ans) 
     float data[];
     unsigned long n;
     float respns[];
     unsigned long m;
     int isign;
     float ans[];
{
    unsigned long i, no2;
    float dum, mag2, *fft;

#ifdef DEBUG
 printf("In convlv > data=%x, n=%d, respns=%x, m=%d, isign=%d, ans=%x ...\n",
        data,n,respns,m,isign,ans);
#endif

    fft = vector(1, n << 1);

#if 0   /* Already done in main() */
    /* Put respns in array of length n. */
    for (i=1; i <= (m-1)/2; i++) {
      respns[n+1-i] = respns[m+1-i];
    }

    /* Pad with zeros */
    for (i=(m+3)/2; i <= n-(m-1)/2; i++) {  
      respns[i] = 0.0;
    }
#endif

#ifdef DEBUG
 printf("In convlv > Calling twofft() ...\n");
#endif
    twofft(data, respns, fft, ans, n);   /* FFT both at once */
    no2 = n >> 1;
    for (i=2; i <= n+2; i += 2) {
      if (isign == 1) {
	/* multiply FFTs to convolve */
#ifdef DEBUG
 printf("In convlv > Convolving by multiplying FFTs\n");
#endif
	ans[i-1] = (fft[i-1]*(dum = ans[i-1]) - fft[i]*ans[i])/no2;
	ans[i] = (fft[i]*dum + fft[i-1]*ans[i])/no2;
      } else {
	if (isign == -1) {
	  if ((mag2 = SQR(ans[i-1])+SQR(ans[i])) == 0.0) {
	    nrerror("Deconvolving at response zero in convlv");
	  }
	  /* divide FFTs to deconvolve */
	  ans[i-1] = (fft[i-1]*(dum = ans[i-1]) + fft[i]*ans[i])/mag2/no2;
	  ans[i] = (fft[i]*dum - fft[i-1]*ans[i])/mag2/no2;
        } else {
	  nrerror("No meaning for isign in convlv");
	}
      } 
    }  
    ans[2] = ans[n+1];  /* Pack last element with first for realft */
#ifdef DEBUG
 printf("In convlv > Calling realft() ...\n");
#endif
    realft(ans, n, -1); /* Inverse transform back to time domain */
    free_vector(fft, 1, n << 1);
#ifdef DEBUG
 printf("In convlv > exiting ... \n");
#endif
}


/*
 *-----------< MAIN >----------------------------
 */

#ifndef TEST_CONVOLVE

main(argc, argv)
  int argc;
  char *argv[];
{
   FILE *fd;
   FILE *outfd;
   float *signal;
   float *response;  
   float *answer;
   int times;         /* times to convolve */
   float meanIndex;   /* index mapping to mean rate */
   int Mhalf;         /* negative part of response signal */
   int totalPts;      /* total no. of pts in pdf */ 
   int N;             /* adjusted size of signal file */
   int powTwo;        /* power of 2 to adjust N to */
   int i;
   float rate, prob;
   int count;
 
   if (argc < 4) {
      printf("Usage: convolve [infile] [outfile] [times]\n");
      exit(1);
   }

  /*
   * Read in the desired file containing signal to 
   * convolve.
   */
#ifdef DEBUG
 printf("In convolve > opening files ... \n");
#endif

   fd = fopen(argv[1], "r");
   if (fd == NULL) {
      printf("Error opening input file\n");
      exit(1);
   }
   outfd = fopen(argv[2], "w");
   if (outfd == NULL) {
      printf("Error opening output file\n");
      exit(1);
   }

  /*
   * Get size of pdf. Allocate signal array and load.
   * Also, calculate mean index. Note: in order not to
   * require storing the rate (abscissa), we implicitly
   * map the rate to the array index and take means, etc.
   * in the array index instead.
   */
#ifdef DEBUG
 printf("In convolve > allocating and loading signal\n");
#endif
   totalPts = 0;
   while (!feof(fd)) {
      fscanf(fd, "%f %f %d\n", &rate, &prob, &count);
      totalPts++;
   } 
   signal = (float *) calloc((totalPts+1)*sizeof(float));
   if (signal == NULL) {
      printf("Error in malloc\n");
      exit(1);
   }
   fseek(fd,0,0);
   i = 1;
   while (!feof(fd)) {
      fscanf(fd,"%f %f %d\n", &rate, &prob, &count);
      signal[i] = prob; 
      i++;
   } 

  /* 
   * Get number of times to convolve. For each ...
   */
   times = atoi(argv[3]);

   while (times) {
      times--;
      Mhalf = 0;
      meanIndex = 0.0;
     /*
      * Find the mean.
      * Find number of points, M, from start to mean.
      */
      for (i=1; i <= totalPts; i++) {
         meanIndex += i*signal[i];
      }
      Mhalf = irint(ceil(meanIndex));  /* index of signal[] >= mean */ 

     /* 
      * Increase length of signal file from K to N 
      * where N >= K+M and N is a power of 2. 
      * K = totalPts; M = max{Mhalf, totalPts-Mhalf}.
      */
      N = (Mhalf > (totalPts - Mhalf)) ? (totalPts + Mhalf) :
          (totalPts*2 - Mhalf);
      powTwo = 1;
      while (powTwo < N) {
         powTwo *= 2; 
      }
      N = powTwo;     /* round N to nearest power of 2 */

     /*
      * Extend size of signal buffer to N
      */
#ifdef DEBUG
 printf("In convolve > reallocating signal ...\n");
#endif

      realloc(signal, N+1);  /* recall: we don't use signal[0] */
      for (i=totalPts+1; i <= N; i++) {
         signal[i] = 0.0;    /* pad with zeros */
      }  

     /* 
      * Allocate buffers for response, and answer arrays.
      */
#ifdef DEBUG
 printf("In convolve > allocating response and answer ...\n");
#endif
      if (response) free(response);
      if (answer) free(answer);
      response = (float *) calloc((N+1)*sizeof(float)); 
      answer = (float *) calloc(((2*N)+1)*sizeof(float)); 
      if ((response == NULL) || (answer == NULL)) {
         printf("Memory allocation error\n");
         exit(1);
      }

     /* 
      * Construct wrap-around version of the signal data 
      * centered around the mean. This will act as the
      * response signal.
      */
#ifdef DEBUG
 printf("In convolve > creating wrap-around ...\n");
#endif

      /* copy positive segment to start */
      for (i=Mhalf; i <= totalPts; i++) {   
        response[i-Mhalf+1] = signal[i];
      }

      /* copy negative segment to end */
      for (i=Mhalf-1; i >= 1; i--) {
        response[N-(Mhalf-1)+i] = signal[i];
      }

#ifdef DEBUG
 printf("In convolve > Calling convlv() ...\n");
#endif
      convlv(signal, N, response, N, 1, answer);  /* do convolution */
      totalPts = N;                               /* new points count */
#ifdef DEBUG
 printf("In convolve > copying answer to signal ...\n");
#endif
      for (i=1; i <= N; i++) {   /* If convolving more than once make */
         signal[i] = answer[i];  /* answer the new signal form. */  
      }
   } /* end while */

  /*
   * Output to file
   */
#ifdef DEBUG
 printf("In convolve > outputting to file ...\n");
#endif
   for (i=1; i <= N; i++) {
      fprintf(outfd, "%d, %.4f\n", i, signal[i]); 
   }
   fclose(fd);
   fclose(outfd);
#ifdef DEBUG
 printf("In convolve > exiting ...\n");
#endif
}

#else TEST_CONVOLVE

main()
{
   FILE *fd;
   float *signal;
   float *response;  
   float *answer;
   int i, N;
   int totalPts;
   int Mhalf;

   N = 1024;
   totalPts = 400;
   if (signal) free(signal);
   signal = (float *) calloc((N+1)*sizeof(float));
   for (i=1; i <= totalPts; i++) {
      signal[i] = 1.5;
   }

   realft(signal, N, 1);
   realft(signal, N, -1);

   for (i=1; i <= N; i++) {
      printf("%.4f %.4f\n", signal[i], signal[i+1]);
   }



#if 0   /*#######*/
   N = 1024;
   totalPts = 400;
   if (response) free(response);
   if (answer) free(answer);
   if (signal) free(signal);
   signal = (float *) calloc((N+1)*sizeof(float));
   response = (float *) calloc((N+1)*sizeof(float)); 
   answer = (float *) calloc(((2*N)+1)*sizeof(float)); 
   if ((response == NULL) || (answer == NULL) || (signal == NULL)) {
      printf("Memory allocation error\n");
      exit(1);
   }

   for (i=1; i <= totalPts; i++) {
      signal[i] = 1.5;
   }
   Mhalf = 200;
   for (i=Mhalf; i <= totalPts; i++) {   
      response[i-Mhalf+1] = 1.0;
   }

   for (i=Mhalf-1; i >= 1; i--) {
      response[N-(Mhalf-1)+i] = 1.0;
   }

   for (i=1; i <= N; i++) {
      printf("%d %.4f\n", i, response[i]);
    }

   convlv(signal, N, response, N, 1, answer);  /* do convolution */
   fd = fopen("test.out", "w");
   if (!fd) {
      printf("Error opening file\n");
   }
   for (i=1; i <= N; i++) {
      fprintf(fd, "%d, %.4f\n", i, answer[i]); 
   }
#endif /*#####*/
}

#endif TEST_CONVOLVE
























